#  Copyright (c) 2021 zfit

import functools
from typing import Any

import tensorflow as tf

import zfit.z.numpy as znp

from ..settings import ztypes
from ..util.deprecation import deprecated
from .tools import _auto_upcast


def exp(x):
    return _auto_upcast(znp.exp(x=x))


@functools.wraps(tf.convert_to_tensor)
def convert_to_tensor(value, dtype=ztypes.float, name=None, preferred_dtype=ztypes.float):
    value = tf.cast(value, dtype=dtype)
    return tf.convert_to_tensor(value=value, dtype=dtype, name=name, dtype_hint=preferred_dtype)


@deprecated(None, "Use z.random.normal instead.")
def random_normal(shape, mean=0.0, stddev=1.0, dtype=ztypes.float, seed=None, name=None):
    return tf.random.normal(shape=shape, mean=mean, stddev=stddev, dtype=dtype, seed=seed, name=name)


@deprecated(None, "Use z.random.uniform instead.")
def random_uniform(shape, minval=0, maxval=None, dtype=ztypes.float, seed=None, name=None):
    return tf.random.uniform(shape=shape, minval=minval, maxval=maxval, dtype=dtype, seed=seed, name=name)


@deprecated(None, "Use z.random.poisson instead.")
def random_poisson(lam: Any, shape: Any, dtype: tf.DType = ztypes.float, seed: Any = None, name: Any = None):
    return tf.random.poisson(lam=lam, shape=shape, dtype=dtype, seed=seed, name=name)


def square(x, name=None):
    return _auto_upcast(tf.square(x, name))


def sqrt(x, name=None):
    return _auto_upcast(tf.sqrt(x, name=name))


def pow(x, y, name=None):
    return _auto_upcast(tf.pow(x, y, name=name))


def complex(real, imag, name=None):
    real = _auto_upcast(real)
    imag = _auto_upcast(imag)
    return _auto_upcast(tf.complex(real=real, imag=imag, name=name))


def check_numerics(tensor: Any, message: Any, name: Any = None):
    """Check whether a tensor is finite and not NaN. Extends TF by accepting complex types as well.

    Args:
        tensor:
        message:
        name:

    Returns:
    """
    if tf.as_dtype(tensor.dtype).is_complex:
        real_check = tf.debugging.check_numerics(tensor=znp.real(tensor), message=message, name=name)
        imag_check = tf.debugging.check_numerics(tensor=znp.imag(tensor), message=message, name=name)
        check_op = tf.group(real_check, imag_check)
    else:
        check_op = tf.debugging.check_numerics(tensor=tensor, message=message, name=name)
    return check_op


reduce_sum = znp.sum

reduce_prod = znp.prod



#  Copyright (c) 2021 zfit

from typing import Any, Callable, Iterable, Union

import tensorflow as tf


def convert_to_container(value: Any, container: Callable = list, non_containers=None,
                         convert_none=False) -> Union[None, Iterable]:
    """Convert `value` into a `container` storing `value` if `value` is not yet a python container.

    Args:
        value:
        container: Converts a tuple to a container.
        non_containers: Types that do not count as a container. Has to
            be a list of types. As an example, if `non_containers` is [list, tuple] and the value
            is [5, 3] (-> a list with two entries),this won't be converted to the `container` but end
            up as (if the container is e.g. a tuple): ([5, 3],) (a tuple with one entry).

    Returns:
    """
    from ..core.interfaces import ZfitData  # here due to dependency
    from ..core.interfaces import ZfitLoss, ZfitModel, ZfitParameter, ZfitSpace
    if non_containers is None:
        non_containers = []
    if not isinstance(non_containers, list):
        raise TypeError("`non_containers` have to be a list or a tuple")
    if value is None and not convert_none:
        return value
    if not isinstance(value, container):
        try:
            non_containers.extend([str, tf.Tensor, ZfitData, ZfitLoss, ZfitModel, ZfitSpace, ZfitParameter])
            if isinstance(value, tuple(non_containers)):
                raise TypeError
            value = container(value)
        except TypeError:
            value = container((value,))
    return value


def is_container(obj):
    """Check if `object` is a list or a tuple.

    Args:
        obj:

    Returns:
        True if it is a *container*, otherwise False
    """
    return isinstance(obj, (list, tuple))



#  Copyright (c) 2021 zfit

import copy

import pytest

import zfit


@pytest.mark.first  # needs to run before initialization
@pytest.mark.parametrize(["n_cpu", 'taken', 'left'], [[3, 5, 0], [0, -1, 0], [10, 3, 7],
                                                      [5, -1, 0], [8, -3, 2], [5, 0, 5]])
def test_cpu_management(n_cpu, taken, left):
    zfit.run.set_n_cpu(n_cpu=n_cpu)
    _cpu = copy.deepcopy(zfit.run._cpu)
    assert zfit.run.n_cpu == n_cpu
    with zfit.run.aquire_cpu(max_cpu=taken) as cpus:
        assert zfit.run.n_cpu == left
        assert isinstance(cpus, list)
        assert len(cpus) == n_cpu - left
    assert zfit.run.n_cpu == n_cpu
    assert _cpu == zfit.run._cpu



#  Copyright (c) 2021 zfit


from zfit.util.temporary import TemporarilySet


def test_simple_x():
    class SimpleX:

        def __init__(self):
            self.x = None

        def _set_x(self, x):
            self.x = x

        def get_x(self):
            return self.x

        def set_x(self, x):
            return TemporarilySet(value=x, setter=self._set_x, getter=self.get_x)

    simple_x = SimpleX()
    simple_x.set_x(42)
    assert simple_x.get_x() == 42
    list1 = [1, 5]
    with simple_x.set_x(list1) as value:
        assert value == list1
        assert simple_x.get_x() == list1
    assert simple_x.get_x() == 42



#  Copyright (c) 2021 zfit

import copy
import random

import numpy as np
import pytest

import zfit
from zfit.core.space import Space, convert_to_space
from zfit.util.exception import (CoordinatesUnderdefinedError,
                                 LimitsIncompatibleError,
                                 LimitsUnderdefinedError,
                                 ShapeIncompatibleError)

lower11 = 1
lower12 = 4
upper11 = 3
upper12 = 7
limit11 = lower11, upper11
limit12 = lower12, upper12
limit11_true = copy.deepcopy(limit11)
limit12_true = copy.deepcopy(limit12)
limit11_area = 2
limit12_area = 3
limit11_axes = (1,)
limit12_axes = (1,)
limit11_obs = ("obs1",)
limit12_obs = ("obs1",)
limit11_axes_true = limit11_axes
limit12_axes_true = limit12_axes
space11 = Space(limits=limit11, axes=limit11_axes)
space12 = Space(limits=limit12, axes=limit12_axes)
space1 = space11 + space12
space11_obs = Space(obs=limit11_obs, limits=limit11)
space12_obs = Space(obs=limit12_obs, limits=limit12)
space1_obs = space11_obs + space12_obs
# arguments1 = (space1, lower1, upper1, limit1_true, limit1_axes, limit1_areas, 2)
arguments1 = []

lower2 = (1, 2, 3)
upper2 = (2, 4, 6)
sub_lower2 = (1, 3)
sub_upper2 = (2, 6)
limit2 = lower2, upper2
sub_limit2 = sub_lower2, sub_upper2
limit2_areas = (6, 18)
sub_limit2_areas = (3, 1.8)
limit2_axes = (1, 5, 6)
sub_limit2_axes = (1, 6)
limit2_obs = ('obs1', 'obs2', 'obs3')
sub_limit2_obs = ('obs1', 'obs3')

space2 = Space(limits=limit2, axes=limit2_axes)
space2_obs = Space(obs=limit2_obs, limits=limit2)
sub_space2 = Space(limits=sub_limit2, axes=sub_limit2_axes)
space2_subbed_axes = space2.get_subspace(axes=sub_limit2_axes)

arguments2 = (space2, lower2, upper2, limit2, limit2_axes, limit2_areas, 2)
sub_arguments2 = (sub_space2, sub_lower2, sub_upper2, sub_limit2, sub_limit2_axes, sub_limit2_areas, 2)


@pytest.mark.parametrize("space1, space2", [
    [space2_subbed_axes, sub_space2]
])
def test_equality(space1, space2):
    """
    Args:
        space1:
        space2:
    """
    assert space1.axes == space2.axes
    assert space1.obs == space2.obs
    np.testing.assert_allclose(space1.rect_limits, space2.rect_limits)
    # TODO: reactivate below
    assert zfit.run(space1.rect_area()) == pytest.approx(zfit.run(space2.rect_area()), rel=1e-8)


# @pytest.mark.skip  # eq missing
def test_sub_space():
    sub_space2_true_axes = Space(axes=sub_limit2_axes, limits=sub_limit2)
    assert sub_space2_true_axes == space2_subbed_axes

    sub_space2_true = Space(obs=sub_limit2_obs, limits=sub_limit2)
    space2_subbed = space2_obs.get_subspace(obs=sub_limit2_obs)
    assert space2_subbed == sub_space2_true


@pytest.mark.skip
@pytest.mark.parametrize("space,lower, upper, limit, axes, areas, n_limits",
                         [
                             # arguments1,
                             # arguments2,
                             # sub_arguments2,
                         ])
def test_space(space, lower, upper, limit, axes, areas, n_limits):
    """
    Args:
        space:
        lower:
        upper:
        limit:
        axes:
        areas:
        n_limits:
    """
    assert space.rect_area() == pytest.approx(sum(areas), rel=1e-8)
    # assert space.iter_areas() == pytest.approx(areas, rel=1e-8)
    # assert sum(space.iter_areas(rel=True)) == pytest.approx(1, rel=1e-7)

    assert space.axes == axes

    assert space.limits == limit

    # iter1_limits1, iter2_limits1 = space.iter_limits(as_tuple=True)
    iter1_limits1_space, iter2_limits1_space = space
    # assert iter1_limits1 == (lower[0], upper[0])
    # assert iter2_limits1 == (lower[1], upper[1])
    # assert iter1_limits1 == (iter1_limits1_space.lower[0], iter1_limits1_space.upper[0])
    # assert iter2_limits1 == (iter2_limits1_space.lower[0], iter2_limits1_space.upper[0])
    # assert space.n_limits == n_limits
    # assert iter1_limits1_space.n_limits == 1
    # assert iter2_limits1_space.n_limits == 1


@pytest.mark.parametrize("space,obs",
                         [
                             # (space1_obs, limit11_obs),
                             (space2_obs, limit2_obs)

                         ])
def test_setting_axes(space, obs):
    """
    Args:
        space:
        obs:
    """
    lower, upper = space.rect_limits
    axes = space.axes
    new_obs = list(copy.deepcopy(obs))
    while len(obs) > 1 and new_obs == list(obs):
        random.shuffle(new_obs)
    new_obs = tuple(new_obs)
    true_lower = np.array(tuple(tuple(low[obs.index(o)] for o in new_obs) for low in lower))
    true_upper = np.array(tuple(tuple(up[obs.index(o)] for o in new_obs) for up in upper))
    new_axes = tuple(range(len(new_obs)))
    coords = Space(obs=new_obs, axes=new_axes)
    # obs_axes = OrderedDict((o, ax) for o, ax in zip(new_obs, new_axes))

    if len(obs) > 1:
        # make sure it was shuffled
        assert new_obs != obs
        assert new_axes != axes
        assert np.any(true_lower != lower)
        assert np.any(true_upper != upper)

    new_space = space.with_coords(coords)
    # check new object
    assert new_axes == new_space.axes
    assert new_obs == new_space.obs
    assert np.all(true_lower == new_space.lower)
    assert np.all(true_upper == new_space.upper)

    # check that old object didn't change
    assert axes == space.axes
    assert obs == space.obs
    assert np.all(lower == space.lower)
    assert np.all(upper == space.upper)

    assert axes == space.axes
    assert obs == space.obs
    assert np.all(lower == space.lower)
    assert np.all(upper == space.upper)


def test_exception():
    invalid_obs = (1, 4)
    with pytest.raises(TypeError):
        Space(obs=invalid_obs)
    with pytest.raises(CoordinatesUnderdefinedError):
        Space(obs=None, limits=limit2)
    with pytest.raises(CoordinatesUnderdefinedError):
        Space(axes=None, limits=limit2)
    with pytest.raises(ShapeIncompatibleError):  # one obs only, but two dims
        Space(obs='obs1', limits=(((1, 2),), ((2, 3),)))
    with pytest.raises(ShapeIncompatibleError):  # two obs but only 1 dim
        Space(obs=['obs1', 'obs2'], limits=(((1,),), ((2,),)))
    with pytest.raises(ShapeIncompatibleError):  # one axis but two dims
        Space(axes=(1,), limits=(((1, 2),), ((2, 3),)))
    with pytest.raises(ShapeIncompatibleError):  # two axes but only 1 dim
        Space(axes=(1, 2), limits=(((1,),), ((2,),)))
    # with pytest.raises(ValueError):  # two obs, two limits but each one only 1 dim
    #     Space(obs=['obs1', 'obs2'], limits=(((1,), (2,)), ((2,), (3,))))


def test_dimensions():
    lower1, lower2 = 1, 2
    upper1, upper2 = 2, 3
    space = Space(obs=['obs1', 'obs2'], limits=(((lower1, lower2),), ((upper1, upper2),)))
    assert space.n_obs == 2
    assert space.n_limits == 1
    low1, low2, up1, up2 = space.limit2d
    assert low1 == lower1
    assert low2 == lower2
    assert up1 == upper1
    assert up2 == upper2

    with pytest.raises(RuntimeError):
        space.limit1d

    with pytest.raises(LimitsIncompatibleError):
        _ = Space(obs='obs1', limits=(((1,), (2,)), ((2,), (3,))))

    space = Space(obs=['obs1', 'obs2'], limits=(((1, 5),), ((2, 13),)))
    assert space.n_obs == 2
    with pytest.raises(RuntimeError):
        space.limit1d

    space = Space(obs='obs1', limits=(((1,),), ((2,),)))
    assert space.n_obs == 1
    assert space.n_limits == 1

    lower1 = 1
    upper1 = 2
    space = Space(axes=(1,), limits=(((lower1,),), ((upper1,),)))
    assert space.n_obs == 1
    assert space.n_limits == 1
    lower, upper = space.limit1d
    assert lower == lower1
    assert upper == upper1


def test_autoconvert():
    with pytest.raises(LimitsUnderdefinedError):
        convert_to_space(obs=['obs1', 'obs2'], limits=(((1, 2),), ((2, 3),)))



#  Copyright (c) 2021 zfit

from typing import Type

import pytest


@pytest.fixture
def test_values():
    import numpy as np
    return np.array([3., 11.3, -0.2, -7.82])


mu_true = 1.4
sigma_true = 1.8
low, high = -4.3, 1.9


@pytest.fixture
def obs1():
    import zfit
    return zfit.Space('obs1', (low, high))


def create_gauss1(nameadd=""):
    import zfit
    obs1 = zfit.Space('obs1', (low, high))

    mu = zfit.Parameter("mu" + nameadd, mu_true, mu_true - 2., mu_true + 7.)
    sigma = zfit.Parameter("sigma" + nameadd, sigma_true, sigma_true - 10., sigma_true + 5.)

    gauss_params1 = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obs1, name="gauss_params1")
    return gauss_params1


def create_mu_sigma_true_params():
    import zfit
    mu_true_param = zfit.Parameter('mu_true', mu_true)
    sigma_true_param = zfit.Parameter('sigma_true', sigma_true)
    return mu_true_param, sigma_true_param


def TmpGaussian():
    import zfit
    class TmpGaussian(zfit.pdf.BasePDF):

        def __init__(self, obs, mu, sigma,
                     dtype: Type = zfit.ztypes.float,
                     name: str = "BasePDF", **kwargs):
            params = {'mu': mu, 'sigma': sigma}
            super().__init__(obs, params, dtype, name, **kwargs)

        def _unnormalized_pdf(self, x):
            x = x.unstack_x()
            mu = self.params['mu']
            sigma = self.params['sigma']

            from zfit import z
            return z.exp((-(x - mu) ** 2) / (
                    2 * sigma ** 2))  # non-normalized gaussian

    return TmpGaussian


def true_gaussian_unnorm_func(x):
    import numpy as np
    return np.exp(- (x - mu_true) ** 2 / (2 * sigma_true ** 2))


def true_gaussian_grad(x):
    import numpy as np

    grad_mu = -0.199471140200716 * (2 * mu_true - 2 * x) * np.exp(
        -(-mu_true + x) ** 2 / (2 * sigma_true ** 2)) / sigma_true ** 3
    grad_sigma = -0.398942280401433 * np.exp(
        -(-mu_true + x) ** 2 / (2 * sigma_true ** 2)) / sigma_true ** 2 + 0.398942280401433 * (
            -mu_true + x) ** 2 * np.exp(-(-mu_true + x) ** 2 / (2 * sigma_true ** 2)) / sigma_true ** 4
    return np.array((grad_mu, grad_sigma)).transpose()


def create_mu_sigma_2(nameadd=""):
    import zfit
    mu2 = zfit.Parameter("mu2" + nameadd, mu_true, mu_true - 2., mu_true + 7.)
    sigma2 = zfit.Parameter("sigma2" + nameadd, sigma_true, sigma_true - 10., sigma_true + 5.)
    return mu2, sigma2


def create_wrapped_gauss(nameadd=""):
    import tensorflow_probability as tfp

    import zfit

    obs1 = zfit.Space('obs1', (low, high))
    mu2, sigma2 = create_mu_sigma_2(nameadd)
    gauss_params = dict(loc=mu2, scale=sigma2)
    tf_gauss = tfp.distributions.Normal
    return zfit.models.dist_tfp.WrapDistribution(tf_gauss, dist_params=gauss_params, obs=obs1, name="tf_gauss1")


def create_gauss3(nameadd=""):
    import zfit
    obs1 = zfit.Space('obs1', (low, high))

    mu3 = zfit.Parameter("mu3" + nameadd, mu_true, mu_true - 2., mu_true + 7.)
    sigma3 = zfit.Parameter("sigma3" + nameadd, sigma_true, sigma_true - 10., sigma_true + 5.)
    gauss3 = zfit.pdf.Gauss(mu=mu3, sigma=sigma3, obs=obs1)
    return gauss3


def create_test_gauss1():
    import zfit
    obs1 = zfit.Space('obs1', (low, high))
    mu, sigma = create_mu_sigma_true_params()
    return TmpGaussian()(name="test_gauss1", mu=mu, sigma=sigma, obs=obs1)


def create_wrapped_normal1(nameadd=""):
    import zfit
    obs1 = zfit.Space('obs1', (low, high))

    mu2, sigma2 = create_mu_sigma_2(nameadd)
    return zfit.pdf.Gauss(mu=mu2, sigma=sigma2, obs=obs1, name='wrapped_normal1')


def create_gaussian_dists():
    return [create_test_gauss1(), create_gauss1("dists")]


# starting tests
# ===============================
@pytest.mark.skip  # TODO: should we have gradients in PDFs? Nope, remove
def test_gradient():
    import numpy as np
    gauss3 = create_gauss3()
    random_vals = np.random.normal(4., 2., size=5)
    tensor_grad = gauss3.gradient(x=random_vals, params=['mu', 'sigma'], norm_range=(-np.infty, np.infty))
    random_vals_eval = tensor_grad.numpy()
    np.testing.assert_allclose(random_vals_eval, true_gaussian_grad(random_vals), rtol=1e-3)


def test_input_space():
    import zfit
    from zfit.util.exception import ObsIncompatibleError
    gauss3 = create_gauss3()
    space = zfit.Space('nonexisting_obs', (-3, 5))
    with pytest.raises(ObsIncompatibleError):
        gauss3.set_norm_range(space)


def test_func(obs1, test_values):
    import numpy as np

    import zfit

    test_values = np.array([3., 11.3, -0.2, -7.82])
    test_values = zfit.Data.from_numpy(obs=obs1, array=test_values)

    gauss_params1 = create_gauss1()

    limits = (-15, 5)

    gauss_func = gauss_params1.as_func(norm_range=limits)
    vals = gauss_func.func(test_values)
    vals_pdf = gauss_params1.pdf(x=test_values, norm_range=limits)
    vals, vals_pdf = [vals.numpy(), vals_pdf.numpy()]
    np.testing.assert_allclose(vals_pdf, vals, rtol=1e-3)  # better assertion?


@pytest.mark.parametrize('pdf_factory',
                         [lambda: create_gaussian_dists()[0],
                          lambda: create_gaussian_dists()[1],
                          lambda: create_wrapped_gauss(),
                          create_wrapped_normal1])
def test_normalization(obs1, pdf_factory):
    import numpy as np
    import tensorflow as tf

    import zfit
    from zfit import z

    test_yield = 1524.3
    dist = pdf_factory()
    samples = tf.cast(np.random.uniform(low=low, high=high, size=100000), dtype=tf.float64)
    small_samples = tf.cast(np.random.uniform(low=low, high=high, size=10), dtype=tf.float64)
    with dist.set_norm_range(zfit.Space(obs1, limits=(low, high))):
        probs = dist.pdf(samples)
        probs_small = dist.pdf(small_samples)
        log_probs = dist.log_pdf(small_samples)
        probs, log_probs = zfit.run(probs, log_probs)
        probs = np.average(probs) * (high - low)
        assert probs == pytest.approx(1., rel=0.05)
        assert log_probs == pytest.approx(tf.math.log(probs_small).numpy(), rel=0.05)
        dist = dist.create_extended(z.constant(test_yield))
        probs = dist.pdf(samples)
        probs_extended = dist.ext_pdf(samples)
        result = probs.numpy()
        result = np.average(result) * (high - low)
        result_ext = np.average(probs_extended) * (high - low)
        assert result == pytest.approx(1, rel=0.05)
        assert result_ext == pytest.approx(test_yield, rel=0.05)


@pytest.mark.parametrize('gauss_factory', [create_gauss1, create_test_gauss1])
def test_sampling_simple(gauss_factory):
    import numpy as np

    gauss = gauss_factory()
    n_draws = 1000
    sample_tensor = gauss.sample(n=n_draws, limits=(low, high))
    sampled_from_gauss1 = sample_tensor.numpy()
    assert max(sampled_from_gauss1[:, 0]) <= high
    assert min(sampled_from_gauss1[:, 0]) >= low
    assert n_draws == len(sampled_from_gauss1[:, 0])
    if gauss.params:
        mu_true1 = gauss.params['mu'].numpy()
        sigma_true1 = gauss.params['sigma'].numpy()
    else:
        mu_true1 = mu_true
        sigma_true1 = sigma_true
    sampled_gauss1_full = gauss.sample(n=10000,
                                       limits=(mu_true1 - abs(sigma_true1) * 3, mu_true1 + abs(sigma_true1) * 3))
    sampled_gauss1_full = sampled_gauss1_full.numpy()
    mu_sampled = np.mean(sampled_gauss1_full)
    sigma_sampled = np.std(sampled_gauss1_full)
    assert mu_sampled == pytest.approx(mu_true1, rel=0.07)
    assert sigma_sampled == pytest.approx(sigma_true1, rel=0.07)


def test_sampling_multiple_limits(obs1):
    import numpy as np

    import zfit

    gauss_params1 = create_gauss1()
    n_draws = 1000
    low1, up1 = -1, 0
    lower_interval = zfit.Space(obs=obs1, limits=(low1, up1))
    low2, up2 = 1, 2
    upper_interval = zfit.Space(obs=obs1, limits=(low2, up2))
    sample_tensor = gauss_params1.sample(n=n_draws, limits=lower_interval + upper_interval)
    sampled_from_gauss1 = sample_tensor.numpy()
    between_samples = np.logical_and(sampled_from_gauss1 < up1, sampled_from_gauss1 > low2)
    assert not any(between_samples)
    assert max(sampled_from_gauss1[:, 0]) <= up2
    assert min(sampled_from_gauss1[:, 0]) >= low1
    assert n_draws == len(sampled_from_gauss1[:, 0])

    mu_true = gauss_params1.params['mu'].numpy()
    sigma_true = gauss_params1.params['sigma'].numpy()
    low1, up1 = mu_true - abs(sigma_true) * 4, mu_true
    lower_interval = zfit.Space(obs=obs1, limits=(low1, up1))
    low2, up2 = mu_true, mu_true + abs(sigma_true) * 4
    upper_interval = zfit.Space(obs=obs1, limits=(low2, up2))
    one_interval = zfit.Space(obs=obs1, limits=(low1, up2))

    sample_tensor5 = gauss_params1.sample(n=10000, limits=lower_interval + upper_interval)
    sampled_gauss1_full = sample_tensor5.numpy()
    mu_sampled = np.mean(sampled_gauss1_full)
    sigma_sampled = np.std(sampled_gauss1_full)
    assert mu_sampled == pytest.approx(mu_true, rel=0.07)
    assert sigma_sampled == pytest.approx(sigma_true, rel=0.07)


def test_analytic_sampling(obs1):
    class SampleGauss(TmpGaussian()):
        pass

    import zfit
    from zfit.core.space import ANY_UPPER
    SampleGauss.register_analytic_integral(func=lambda limits, params, model: 2 * limits.upper[0][0],
                                           limits=zfit.Space(limits=(-float("inf"), ANY_UPPER),
                                                             axes=(0,)))  # DUMMY!
    SampleGauss.register_inverse_analytic_integral(func=lambda x, params: x + 1000.)

    mu, sigma = create_mu_sigma_true_params()
    gauss1 = SampleGauss(obs=obs1, mu=mu, sigma=sigma)
    sample = gauss1.sample(n=10000, limits=(2., 5.))
    sample.set_data_range((1002, 1005))

    sample = sample.numpy()

    assert 1004. <= min(sample[0])
    assert 10010. >= max(sample[0])


def test_multiple_limits(obs1):
    import zfit
    from zfit.core.space import MultiSpace

    gauss_params1 = create_gauss1()
    dims = (0,)
    simple_limits = (-3.2, 9.1)
    multiple_limits_lower = ((-3.2,), (1.1,), (2.1,))
    multiple_limits_upper = ((1.1,), (2.1,), (9.1,))

    multiple_limits_range = MultiSpace(
        [zfit.Space(limits=(low, up), axes=dims) for low, up in zip(multiple_limits_lower, multiple_limits_upper)])
    integral_simp = gauss_params1.integrate(limits=simple_limits, norm_range=False)
    integral_mult = gauss_params1.integrate(limits=multiple_limits_range, norm_range=False)
    integral_simp_num = gauss_params1.numeric_integrate(limits=simple_limits, norm_range=False)
    integral_mult_num = gauss_params1.numeric_integrate(limits=multiple_limits_range, norm_range=False)

    integral_simp, integral_mult = [integral_simp.numpy(), integral_mult.numpy()]
    integral_simp_num, integral_mult_num = [integral_simp_num.numpy(), integral_mult_num.numpy()]
    assert integral_simp == pytest.approx(integral_mult, rel=1e-2)  # big tol as mc is used
    assert integral_simp == pytest.approx(integral_simp_num, rel=1e-2)  # big tol as mc is used
    assert integral_simp_num == pytest.approx(integral_mult_num, rel=1e-2)  # big tol as mc is used


def test_copy(obs1):
    gauss_params1 = create_gauss1()
    new_gauss = gauss_params1.copy()
    assert new_gauss == gauss_params1
    assert new_gauss is not gauss_params1


def test_set_yield(obs1):
    gauss_params1 = create_gauss1()
    assert not gauss_params1.is_extended
    gauss_params1._set_yield(10)
    assert gauss_params1.is_extended
    from zfit.util.exception import AlreadyExtendedPDFError
    with pytest.raises(AlreadyExtendedPDFError):
        gauss_params1._set_yield(15)

    from zfit.util.exception import BreakingAPIChangeError
    with pytest.raises(BreakingAPIChangeError):
        gauss_params1._set_yield(None)


@pytest.mark.flaky(reruns=3)
def test_projection_pdf(test_values):
    import numpy as np

    import zfit

    x = zfit.Space("x", limits=(-1, 1))
    y = zfit.Space("y", limits=(-1, 1))

    def correlated_func(self, x):
        x, y = x.unstack_x()
        value = (((x - y ** 3) ** 2) + 0.1)
        return value

    def correlated_func_integrate_x(y, limits):
        lower, upper = limits.rect_limits

        def integ(x, y):
            return 0.333333333333333 * x ** 3 - 1.0 * x ** 2 * y ** 3 + x * (1.0 * y ** 6 + 0.1)

        return integ(y, upper) - integ(y, lower)

    def correlated_func_integrate_y(x, limits):
        lower, upper = limits.rect_limits

        def integ(x, y):
            return -0.5 * x * y ** 4 + 0.142857142857143 * y ** 7 + y * (1.0 * x ** 2 + 0.1)

        return (integ(x, upper) - integ(x, lower))[0]

    obs = x * y
    from zfit.models.special import SimplePDF
    gauss_xy = SimplePDF(func=correlated_func, obs=obs)
    assert gauss_xy.create_projection_pdf(limits_to_integrate=y).norm_range == x
    proj_pdf = gauss_xy.create_projection_pdf(limits_to_integrate=y)
    test_values = np.array([-0.95603563, -0.84636306, -0.83895759, 2.62608006, 1.02336499,
                            -0.99631608, -1.22185623, 0.83838586, 2.77894762, -2.48259488,
                            1.5440374, 0.1109899, 0.20873491, -2.45271623, 2.04510553,
                            0.31566277, -1.55696965, 0.36304538, 0.77765786, 3.92630088])
    true_probs = correlated_func_integrate_y(test_values, y) / gauss_xy.integrate(limits=obs, norm_range=False)
    probs = proj_pdf.pdf(x=test_values)
    probs = probs.numpy()
    np.testing.assert_allclose(probs, true_probs, rtol=1e-3)  # MC normalization



#  Copyright (c) 2020 zfit

import zfit

# create space
obs = zfit.Space("x", limits=(-10, 10))

# parameters
mu = zfit.Parameter("mu", 1., -4, 6)
sigma = zfit.Parameter("sigma", 1., 0.1, 10)
lambd = zfit.Parameter("lambda", -0.06, -1, -0.01)
frac = zfit.Parameter("fraction", 0.3, 0, 1)

# model building, pdf creation
gauss = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obs)
exponential = zfit.pdf.Exponential(lambd, obs=obs)
model = zfit.pdf.SumPDF([gauss, exponential], fracs=frac)

# data
n_sample = 10000

data = model.create_sampler(n_sample, limits=obs)
data.resample()

# set the values to a start value for the fit
mu.set_value(0.5)
sigma.set_value(1.2)
lambd.set_value(-0.05)
frac.set_value(0.07)

# create NLL
nll = zfit.loss.UnbinnedNLL(model=model, data=data)

# create a minimizer
minimizer = zfit.minimize.Minuit()
result = minimizer.minimize(nll)
print(result)

# do the error calculations, here with minos
param_hesse = result.hesse()
param_errors, new_result = result.errors()
print(result.params)



#  Copyright (c) 2021 zfit

import abc
from typing import Iterable, List, Optional, Set, Tuple, Union

from ordered_set import OrderedSet

from ..core.basemodel import BaseModel
from ..core.dependents import _extract_dependencies
from ..core.dimension import get_same_obs
from ..core.interfaces import ZfitFunctorMixin, ZfitModel, ZfitSpace
from ..core.space import Space, combine_spaces
from ..util import ztyping
from ..util.container import convert_to_container
from ..util.exception import (LimitsIncompatibleError,
                              NormRangeNotSpecifiedError,
                              SpaceIncompatibleError)


def extract_daughter_input_obs(obs: ztyping.ObsTypeInput, spaces: Iterable[ZfitSpace]) -> ZfitSpace:
    """Extract the common space from `spaces` by combining them, test against obs.

    The `obs` are assumed to be the obs given to a functor while the `spaces` are the spaces of the daughters.
    First, the combined space from the daughters is extracted. If no `obs` are given, this is returned.
    If `obs` are given, it is checked whether they agree. If they agree, and no limit is set on `obs` (i.e. they
    are pure strings), the inferred limits are used, sorted by obs. Otherwise, obs is directly used.

    Args:
        obs:
        spaces:

    Returns:
    """
    spaces = convert_to_container(spaces)
    # combine spaces and limits
    try:
        models_space = combine_spaces(*spaces)
    except LimitsIncompatibleError:  # then only add obs
        extracted_obs = _extract_common_obs(obs=tuple(space.obs for space in spaces))
        models_space = Space(obs=extracted_obs)

    if obs is None:
        obs = models_space
    else:
        if isinstance(obs, Space):
            obs = obs
        else:
            obs = Space(obs=obs)
        # if not frozenset(obs.obs) == frozenset(models_space.obs):  # not needed, example projection
        #     raise SpaceIncompatibleError("The given obs do not coincide with the obs from the daughter models.")
        if not obs.obs == models_space.obs and not obs.limits_are_set:
            obs = models_space.with_obs(obs.obs)

    return obs


class FunctorMixin(ZfitFunctorMixin, BaseModel):

    def __init__(self, models, obs, **kwargs):
        models = convert_to_container(models, container=list)
        obs = extract_daughter_input_obs(obs=obs, spaces=[model.space for model in models])

        super().__init__(obs=obs, **kwargs)
        # TODO: needed? remove below
        self._model_obs = tuple(model.obs for model in models)

    def _get_params(self, floating: Optional[bool] = True, is_yield: Optional[bool] = None,
                    extract_independent: Optional[bool] = True) -> Set["ZfitParameter"]:
        params = super()._get_params(floating, is_yield, extract_independent)
        if is_yield is not True:
            params = params.union(*(model.get_params(floating=floating, is_yield=False,
                                                     extract_independent=extract_independent)
                                    for model in self.models))
        return params

    # def _infer_space_from_daughters(self):
    #     space = set(model.space for model in self.models)
    #     obs = set(norm_range.obs for norm_range in space)
    #     if len(space) == 1:
    #         return space.pop()
    #     elif len(obs) > 1:  # TODO(Mayou36, #77): different obs?
    #         return None
    #     else:
    #         return False

    def _get_dependencies(self):
        dependents = super()._get_dependencies()  # get the own parameter dependents
        model_dependents = _extract_dependencies(self.get_models())
        return dependents.union(model_dependents)

    @property
    def models(self) -> List[ZfitModel]:
        """Return the models of this `Functor`.

        Can be `pdfs` or `funcs`.
        """
        return self._models

    @property
    def _model_same_obs(self):
        return get_same_obs(self._model_obs)

    @property
    @abc.abstractmethod
    def _models(self) -> List[ZfitModel]:
        raise NotImplementedError

    def get_models(self, names=None) -> List[ZfitModel]:
        if names is None:
            models = list(self.models)
        else:
            raise ValueError("name not supported currently.")
            # models = [self.models[name] for name in names]
        return models

    def _check_input_norm_range_default(self, norm_range, caller_name="", none_is_error=True):
        if norm_range is None:
            try:
                norm_range = self.norm_range
            except AttributeError:
                raise NormRangeNotSpecifiedError("The normalization range is `None`, no default norm_range is set")
        return self._check_input_norm_range(norm_range=norm_range, none_is_error=none_is_error)


def _extract_common_obs(obs: Tuple[Union[Tuple[str], Space]]) -> Tuple[str]:
    obs_iter = [space.obs if isinstance(space, Space) else space for space in obs]
    unique_obs = []
    for obs in obs_iter:
        for o in obs:
            if o not in unique_obs:
                unique_obs.append(o)
    return tuple(unique_obs)



#  Copyright (c) 2021 zfit
import itertools
import platform
from collections import OrderedDict

import numpy as np
import pytest
import scipy.optimize
from ordered_set import OrderedSet

import zfit.minimizers.optimizers_tf
from zfit.minimizers.base_tf import WrapOptimizer
from zfit.util.exception import OperationNotAllowedError

true_mu = 4.5
true_sigma = 2
true_lambda = -0.03

parameter_tol = 0.1
max_distance_to_min = 2.5


def create_loss(obs1):
    mu_param = zfit.Parameter("mu", true_mu - 2.5, -5., 9.,
                              step_size=0.03
                              )
    sigma_param = zfit.Parameter("sigma", true_sigma * 0.3, 0.01, 10,
                                 step_size=0.03)
    lambda_param = zfit.Parameter("lambda", true_lambda * 0.3, -0.5, -0.0003,
                                  step_size=0.001)

    gauss1 = zfit.pdf.Gauss(mu=mu_param, sigma=sigma_param, obs=obs1)
    exp1 = zfit.pdf.Exponential(lam=lambda_param, obs=obs1)

    sum_pdf1 = zfit.pdf.SumPDF([gauss1, exp1], 0.8)
    # load params for sampling
    with mu_param.set_value(true_mu):
        with sigma_param.set_value(true_sigma):
            with lambda_param.set_value(true_lambda):
                sampled_data = sum_pdf1.create_sampler(n=25000)
                sampled_data.resample()

                loss = zfit.loss.UnbinnedNLL(model=sum_pdf1, data=sampled_data, options={'subtr_const': True})
                minimum = loss.value().numpy()

    return loss, minimum, (mu_param, sigma_param, lambda_param)


verbosity = None


def make_min_grad_hesse():
    minimizers = [
        zfit.minimize.ScipyTruncNCV1,
        # zfit.minimize.ScipyTrustNCGV1,  # Too bad
        # zfit.minimize.ScipyTrustKrylovV1,  # Too bad
        zfit.minimize.ScipySLSQPV1,
        zfit.minimize.ScipyLBFGSBV1,
        zfit.minimize.ScipyTrustConstrV1,
    ]
    min_options = []
    for opt in minimizers:
        grad = opt._VALID_SCIPY_GRADIENT
        hess = opt._VALID_SCIPY_HESSIAN
        if not grad:
            grad = {None}  # the default is None, so this will skip it
        if not hess:
            hess = {None}
        product = itertools.product([opt], grad, hess)
        min_options.extend(product)
    return min_options


@pytest.mark.parametrize('minimizer_gradient_hessian', make_min_grad_hesse())
@pytest.mark.flaky(reruns=3)
def test_scipy_derivative_options(minimizer_gradient_hessian):
    minimizer_cls, gradient, hessian = minimizer_gradient_hessian
    loss, true_min, params = create_loss(obs1=obs1)
    (mu_param, sigma_param, lambda_param) = params
    kwargs = {}

    if gradient is not None:
        kwargs['gradient'] = gradient
    if hessian is not None:
        kwargs['hessian'] = hessian

    try:
        minimizer = minimizer_cls(**kwargs)
    except ValueError as error:  # we test a not allowed combination
        if 'Whenever the gradient is estimated via finite-differences' in error.args[0]:
            return
        else:
            raise

    result = minimizer.minimize(loss=loss)
    assert result.valid

    found_min = loss.value().numpy()
    assert true_min + max_distance_to_min >= found_min

    aval, bval, cval = zfit.run((mu_param, sigma_param, lambda_param))

    assert true_mu == pytest.approx(aval, abs=parameter_tol)
    assert true_sigma == pytest.approx(bval, abs=parameter_tol)
    assert true_lambda == pytest.approx(cval, abs=parameter_tol)
    assert result.converged


minimizers = [
    # minimizers, minimizer_kwargs, do error estimation
    # TensorFlow minimizers
    # (zfit.minimizers.optimizers_tf.WrapOptimizer, dict(optimizer=tf.keras.optimizers.Adam(learning_rate=0.05)),
    #  False),
    (zfit.minimize.Adam, dict(learning_rate=0.05, verbosity=verbosity, tol=0.00001), False),  # works

    # Minuit minimizer
    (zfit.minimize.Minuit, {"verbosity": verbosity}, {'error': True, 'longtests': True}),  # works

    # Ipyopt minimizer

    # TensorFlow Probability minimizer
    # (BFGS, {}, True),  # doesn't work as it uses the graph, violates assumption in minimizer

    # SciPy Minimizer
    (zfit.minimize.ScipyLBFGSBV1, {'tol': 1e-5, "verbosity": verbosity}, {'error': True,
                                                                          'numgrad': False, 'approx': True}),
    # (zfit.minimize.ScipyTrustNCGV1, {'tol': 1e-5, "verbosity": verbosity}, True),
    # (zfit.minimize.ScipyTrustKrylovV1, {"verbosity": verbosity}, True),  # Too unstable
    (zfit.minimize.ScipyTrustConstrV1, {"verbosity": verbosity, }, {'error': True, 'longtests': True}),
    (zfit.minimize.ScipyPowellV1, {"verbosity": verbosity, }, {'error': True}),
    (zfit.minimize.ScipySLSQPV1, {"verbosity": verbosity, }, {'error': True}),
    # (zfit.minimize.ScipyCOBYLAV1, {"verbosity": verbosity, }, {'error': True}),  # Too bad
    # (zfit.minimize.ScipyDoglegV1, {'tol': 1e-5, "verbosity": verbosity}, True),  # works badly
    # (zfit.minimize.ScipyNewtonCGV1, {"verbosity":verbosity,}, {'error': True}),  # Too sensitive? Fails in line-search?
    (zfit.minimize.ScipyTruncNCV1, {"verbosity": verbosity, }, {'error': True}),

    # NLopt minimizer
    (zfit.minimize.NLoptLBFGSV1, {"verbosity": verbosity, }, {'error': True, 'longtests': True}),
    (zfit.minimize.NLoptTruncNewtonV1, {"verbosity": verbosity, }, {'error': True}),
    (zfit.minimize.NLoptSLSQPV1, {"verbosity": verbosity, }, {'error': True}),
    (zfit.minimize.NLoptMMAV1, {"verbosity": verbosity, }, {'error': True}),
    (zfit.minimize.NLoptCCSAQV1, {"verbosity": verbosity, }, {'error': True}),
    (zfit.minimize.NLoptSubplexV1, {"verbosity": verbosity, }, {'error': True}),
    (zfit.minimize.NLoptCOBYLAV1, {"verbosity": verbosity, }, {'error': True}),
    (zfit.minimize.NLoptMLSLV1, {"verbosity": verbosity, }, {'error': True}),
    (zfit.minimize.NLoptStoGOV1, {"verbosity": verbosity, }, {'error': True}),
    (zfit.minimize.NLoptBOBYQAV1, {"verbosity": verbosity, }, {'error': True}),
    (zfit.minimize.NLoptISRESV1, {"verbosity": verbosity, }, {'error': True}),
    (zfit.minimize.NLoptESCHV1, {"verbosity": verbosity, }, {'error': True}),
    (zfit.minimize.NLoptShiftVarV1, {"verbosity": verbosity, }, {'error': True}),

    # (zfit.minimize.Scipy, {'tol': 1e-8, 'algorithm': 'CG'}, False),
    # (zfit.minimize.Scipy, {'tol': 1e-8, 'algorithm': 'BFGS'}, False),  # too bad

    # (zfit.minimize.NLopt, {'tol': 1e-8, 'algorithm': nlopt.LN_NELDERMEAD}, True),  # performs too bad

]

# minimizers = [(zfit.minimize.Minuit, {"verbosity": verbosity, 'gradient': True}, {'error': True, 'longtests': True})]
# minimizers = [(zfit.minimize.IpyoptV1, {'verbosity': 7}, True)]
# minimizers = [(zfit.minimize.ScipyLBFGSBV1, {'verbosity': 7}, True)]
# minimizers = [(zfit.minimize.ScipyPowellV1, {'verbosity': 7}, True)]
# minimizers = [(zfit.minimize.ScipySLSQPV1, {'verbosity': 7}, True)]
# minimizers = [(zfit.minimize.ScipyNelderMeadV1, {'verbosity': 7}, True)]
# minimizers = [(zfit.minimize.ScipyCOBYLAV1, {'verbosity': 7}, True)]
# minimizers = [(zfit.minimize.ScipyNewtonCGV1, {'verbosity': 7}, True)]
# minimizers = [(zfit.minimize.ScipyTrustNCGV1, {'tol': 1e-3, 'verbosity': 7}, True)]
# minimizers = [(zfit.minimize.ScipyTruncNCV1, {'tol': 1e-5, 'verbosity': 7}, True)]
# minimizers = [(zfit.minimize.ScipyDoglegV1, {'tol': 1e3, 'verbosity': 7}, True)]
# minimizers = [(zfit.minimize.ScipyTrustConstrV1, {'tol': 1e-5, 'verbosity': 7}, True)]
# minimizers = [(zfit.minimize.ScipyTrustKrylovV1, {'verbosity': 7}, True)]
# minimizers = [(zfit.minimize.NLoptLBFGSV1, {'verbosity': 7}, {'error': True, 'longtests': True})]
# minimizers = [(zfit.minimize.NLoptTruncNewtonV1, {'verbosity': 7}, True)]
# minimizers = [(zfit.minimize.NLoptSLSQPV1, {'verbosity': 7}, True)]
# minimizers = [(zfit.minimize.NLoptMMAV1, {'verbosity': 7}, True)]
# minimizers = [(zfit.minimize.NLoptCCSAQV1, {'verbosity': 7}, True)]
# minimizers = [(zfit.minimize.NLoptMLSLV1, {'verbosity': 7}, {'error': True, 'longtests': True})]
# minimizers = [(zfit.minimize.NLoptStoGOV1, {'verbosity': 7}, {'error': True, 'longtests': True})]  # DOESN'T WORK!
# minimizers = [(zfit.minimize.NLoptSubplexV1, {'verbosity': 7}, {'error': True, 'longtests': True})]
# minimizers = [(zfit.minimize.NLoptESCHV1, {'verbosity': 7}, {'error': True, 'longtests': True})]
# minimizers = [(zfit.minimize.NLoptISRESV1, {'verbosity': 7}, {'error': True, 'longtests': True})]
# minimizers = [(zfit.minimize.NLoptBOBYQAV1, {'verbosity': 7}, {'error': True, 'longtests': True})]
# minimizers = [(zfit.minimize.NLoptShiftVarV1, {'verbosity': 7, 'rank': 2}, {'error': True, 'longtests': True})]
# minimizers = [(zfit.minimize.Minuit, {'verbosity': 6}, True)]
# minimizers = [(zfit.minimize.BFGS, {'verbosity': 6}, True)]

if not platform.system() == 'Darwin':  # TODO: Ipyopt installation on macosx not working
    minimizers.append((zfit.minimize.IpyoptV1, {"verbosity": verbosity}, {'error': True, 'longtests': True}))
# sort for xdist: https://github.com/pytest-dev/pytest-xdist/issues/432
minimizers = sorted(minimizers, key=lambda val: repr(val))

minimizers_small = [
    (zfit.minimize.NLoptLBFGSV1, {}, True),
    (zfit.minimize.ScipyTrustConstrV1, {}, True),
    (zfit.minimize.Minuit, {}, True),
]
if not platform.system() == 'Darwin':  # TODO: Ipyopt installation on macosx not working
    minimizers_small.append((zfit.minimize.IpyoptV1, {}, False))
# sort for xdist: https://github.com/pytest-dev/pytest-xdist/issues/432
minimizers_small = sorted(minimizers_small, key=lambda val: repr(val))

obs1 = zfit.Space(obs='obs1', limits=(-2.4, 9.1))
obs1_split = (zfit.Space(obs='obs1', limits=(-2.4, 1.3))
              + zfit.Space(obs='obs1', limits=(1.3, 2.1))
              + zfit.Space(obs='obs1', limits=(2.1, 9.1)))


def test_floating_flag():
    obs = zfit.Space("x", limits=(-2, 3))
    mu = zfit.Parameter("mu", 1.2, -4, 6)
    sigma = zfit.Parameter("sigma", 1.3, 0.1, 10)
    sigma.floating = False
    gauss = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obs)
    normal_np = np.random.normal(loc=2., scale=3., size=10000)
    data = zfit.Data.from_numpy(obs=obs, array=normal_np)
    nll = zfit.loss.UnbinnedNLL(model=gauss, data=data)
    minimizer = zfit.minimize.Minuit()
    result = minimizer.minimize(nll, params=[mu, sigma])
    assert list(result.params.keys()) == [mu]
    assert sigma not in result.params


@pytest.mark.parametrize("params", [
    # np.random.normal(size=5),
    [1.4, 0.6, 1.5],
    {'value': [1.4, 0.6, 1.5],
     'lower': np.ones(3) * (-5),
     'upper': np.ones(3) * (9),
     'step_size': np.linspace(0.1, 0.2, 3)}])
@pytest.mark.parametrize("minimizer_class_and_kwargs", minimizers_small)
@pytest.mark.flaky(reruns=1)
def test_minimize_pure_func(params, minimizer_class_and_kwargs):
    zfit.run.set_autograd_mode(False)
    zfit.run.set_graph_mode(False)
    minimizer_class, minimizer_kwargs, _ = minimizer_class_and_kwargs
    minimizer = minimizer_class(**minimizer_kwargs)
    func = scipy.optimize.rosen
    func.errordef = 0.5
    if isinstance(minimizer, WrapOptimizer):
        with pytest.raises(OperationNotAllowedError):
            _ = minimizer.minimize(func, params)
    else:
        result = minimizer.minimize(func, params)
        assert result.valid
    result.hesse(name='hesse_np')
    for param, error in zip(result.params, [0.32, 0.64, 1.3]):
        assert pytest.approx(result.params[param]['hesse_np']['error'], rel=0.15) == error
    param = list(result.params)[1]
    result.errors(param, name='errors')
    assert pytest.approx(result.params[param]['errors']['lower'], rel=0.15) == -0.61
    assert pytest.approx(result.params[param]['errors']['upper'], rel=0.15) == 0.7


def test_dependent_param_extraction():
    obs = zfit.Space("x", limits=(-2, 3))
    mu = zfit.Parameter("mu", 1.2, -4, 6)
    sigma = zfit.Parameter("sigma", 1.3, 0.1, 10)
    sigma1 = zfit.ComposedParameter('sigma1', lambda sigma, mu: sigma + mu,
                                    [sigma, mu])
    gauss = zfit.pdf.Gauss(mu=mu, sigma=sigma1, obs=obs)
    normal_np = np.random.normal(loc=2., scale=3., size=10)
    data = zfit.Data.from_numpy(obs=obs, array=normal_np)
    nll = zfit.loss.UnbinnedNLL(model=gauss, data=data)
    minimizer = zfit.minimize.Minuit()
    params_checked = OrderedSet(minimizer._check_convert_input(nll, params=[mu, sigma1])[1])
    assert {mu, sigma} == set(params_checked)
    sigma.floating = False
    params_checked = minimizer._check_convert_input(nll, params=[mu, sigma1])[1]
    assert {mu, } == set(params_checked)


# @pytest.mark.run(order=4)
# chunksizes = [100000, 3000]
chunksizes = [100000]
# skip the numerical gradient due to memory leak bug, TF2.3 fix: https://github.com/tensorflow/tensorflow/issues/35010
# num_grads = [bo for bo in [False, True] if not bo or zfit.run.get_graph_mode()]
numgrads = [False, True]
# num_grads = [True]
# num_grads = [False]

spaces_all = [obs1, obs1_split]

error_scales = {
    None: 1,
    1: 1,
    2: 2

}


@pytest.mark.parametrize("chunksize", chunksizes)
@pytest.mark.parametrize("numgrad", numgrads)
@pytest.mark.parametrize("spaces", spaces_all)
# @pytest.mark.parametrize("errordef", [0.5, 1.0, 2.25, 4.])
# @pytest.mark.parametrize("cl_scale", [(0.683, 1), (0.9548, 2), (0.99747, 3)])  # cl and expected scale of error
@pytest.mark.parametrize("minimizer_class_and_kwargs", minimizers)
@pytest.mark.flaky(reruns=3)
def test_minimizers(minimizer_class_and_kwargs, chunksize, numgrad, spaces,
                    pytestconfig):
    long_clarg = pytestconfig.getoption("longtests")
    # long_clarg = True
    # zfit.run.chunking.active = True
    # zfit.run.chunking.max_n_points = chunksize
    zfit.run.set_autograd_mode(not numgrad)

    # minimize_func(minimizer_class_and_kwargs, obs=spaces)

    minimizer_class, minimizer_kwargs, test_error = minimizer_class_and_kwargs

    if not isinstance(test_error, dict):
        test_error = {'error': test_error}

    numgrad = test_error.get('numgrad', False)
    do_long = test_error.get('longtests', False)
    has_approx = test_error.get('approx', False)
    test_error = test_error['error']

    skip_tests = (not long_clarg
                  and not do_long
                  and not (chunksize == chunksizes[0]
                           and numgrad == numgrads[0]
                           and spaces is spaces_all[0]))
    if skip_tests:
        return
    if not long_clarg and not do_long:
        test_error = False

    # start actual test
    obs = spaces
    loss, true_min, params = create_loss(obs1=obs)
    (mu_param, sigma_param, lambda_param) = params
    minimizer_hightol = minimizer_class(**{**minimizer_kwargs,
                                           'tol': 5.})

    minimizer = minimizer_class(**minimizer_kwargs)

    # Currently not working, stop the test here. Memory leak?
    from zfit.minimizers.minimizer_tfp import BFGS
    if isinstance(minimizer, (BFGS, WrapOptimizer)) and numgrad:
        return
    init_vals = zfit.run(params)
    result = minimizer.minimize(loss=loss)
    zfit.param.set_values(params, init_vals)
    result_lowtol = minimizer_hightol.minimize(loss=loss)
    zfit.param.set_values(params, init_vals)
    result_lowtol2 = minimizer.minimize(loss=result_lowtol)

    assert result.valid
    assert result_lowtol.valid
    assert result_lowtol2.valid
    found_min = loss.value().numpy()
    assert true_min + max_distance_to_min >= found_min

    assert result_lowtol2.fmin == pytest.approx(result.fmin, abs=2.)
    if not isinstance(minimizer, zfit.minimize.IpyoptV1):
        assert result_lowtol2.info['n_eval'] < 1.2 * result.info['n_eval']  # should not be more, surely not a lot

    aval, bval, cval = (zfit.run(v) for v in
                        (mu_param, sigma_param, lambda_param))

    assert true_mu == pytest.approx(aval, abs=parameter_tol)
    assert true_sigma == pytest.approx(bval, abs=parameter_tol)
    assert true_lambda == pytest.approx(cval, abs=parameter_tol)
    assert result.converged

    # Test Hesse
    if test_error:
        for cl, errscale in [(0.683, 1), (0.9548, 2), (0.99747, 3)]:
            hesse_methods = ['hesse_np']
            profile_methods = ['zfit_error']
            from zfit.minimizers.minimizer_minuit import Minuit
            hesse_methods.append('minuit_hesse')
            profile_methods.append('minuit_minos')
            if isinstance(minimizer, Minuit):
                # TODO: Move up for all once https://github.com/scikit-hep/iminuit/issues/631 fixed
                hesse_methods.append('approx')

            rel_error_tol = 0.15
            for method in hesse_methods:
                name = f'{method}_{cl:.3g}'
                sigma_hesse = result.hesse(params=sigma_param, method=method, name=name, cl=cl)
                assert tuple(sigma_hesse.keys()) == (sigma_param,)
                errors = result.hesse(method=method, name=name, cl=cl)
                sigma_hesse = sigma_hesse[sigma_param]
                can_be_none = method == "approx" and not has_approx
                # skip if it can be None and it is None
                sigma_error_true = 0.015 * errscale
                if not (can_be_none and errors[sigma_param].get('error') is None):
                    assert abs(errors[sigma_param]['error']) == pytest.approx(sigma_error_true,
                                                                              abs=rel_error_tol)
                if not (can_be_none and errors[lambda_param].get('error') is None):
                    assert abs(errors[lambda_param]['error']) == pytest.approx(0.01 * errscale,
                                                                               abs=0.01)
                if not (can_be_none and sigma_hesse.get('error') is None):
                    assert abs(sigma_hesse['error']) == pytest.approx(sigma_error_true, abs=rel_error_tol)

            for profile_method in profile_methods:
                # Test Error
                pname = f'{profile_method}_{cl:.3g}'

                a_errors, _ = result.errors(params=mu_param, method=profile_method, name=pname, cl=cl)
                assert tuple(a_errors.keys()) == (mu_param,)
                errors, _ = result.errors(method=profile_method, name=pname, cl=cl)
                a_error = a_errors[mu_param]
                assert a_error["lower"] == pytest.approx(-a_error['upper'],
                                                         rel=0.1)
                assert a_error["lower"] == pytest.approx(-0.021 * errscale, rel=rel_error_tol)
                assert errors[sigma_param]["lower"] == pytest.approx(-sigma_error_true,
                                                                     rel=rel_error_tol)
                assert errors[lambda_param]['lower'] == pytest.approx(-0.007 * errscale,
                                                                      rel=rel_error_tol)
                assert errors[lambda_param]['upper'] == pytest.approx(0.007 * errscale,
                                                                      rel=rel_error_tol)

                assert errors[mu_param]['lower'] == pytest.approx(a_error['lower'],
                                                                  rel=rel_error_tol)
                assert errors[mu_param]['upper'] == pytest.approx(a_error['upper'],
                                                                  rel=rel_error_tol)

                # Test Error method name
                a_errors, _ = result.errors(params=mu_param, method=profile_method,
                                            name='error1')
                assert tuple(a_errors.keys()) == (mu_param,)
                errors, _ = result.errors(name='error42',
                                          method=profile_method)
                a_error = a_errors[mu_param]

                assert a_error['lower'] == pytest.approx(
                    result.params[mu_param]['error42']['lower'], rel=rel_error_tol)
                assert a_error['lower'] == pytest.approx(
                    result.params[mu_param]['error1']['lower'], rel=rel_error_tol)
                for param, errors2 in result.params.items():
                    assert errors[param]['lower'] == pytest.approx(
                        errors2['error42']['lower'], rel=rel_error_tol)
                    assert errors[param]['upper'] == pytest.approx(
                        errors2['error42']['upper'], rel=rel_error_tol)

                # test custom error
                def custom_error_func(result, params, cl):
                    return OrderedDict(
                        (param, {'myval': 42}) for param in params), None

                custom_errors, _ = result.errors(method=custom_error_func,
                                                 name='custom_method1')
                for param, errors2 in result.params.items():
                    assert custom_errors[param]['myval'] == 42



#  Copyright (c) 2021 zfit

from typing import Dict, Iterable, List, Set, Union

import numpy as np

import zfit
from zfit.util.exception import SpaceIncompatibleError

from ..util import ztyping
from ..util.container import convert_to_container
from .interfaces import ZfitDimensional


class BaseDimensional(ZfitDimensional):

    def _check_n_obs(self, space):
        if self._N_OBS is not None:
            if len(space.obs) != self._N_OBS:
                raise SpaceIncompatibleError(f"Exactly {self._N_OBS} obs are allowed, {space.obs} are given.")

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if not hasattr(cls, "_N_OBS"):
            cls._N_OBS = None

    @property
    def obs(self) -> ztyping.ObsTypeReturn:
        return self.space.obs

    @property
    def axes(self) -> ztyping.AxesTypeReturn:
        return self.space.axes

    @property
    def n_obs(self) -> int:
        return self.space.n_obs


def get_same_obs(obs):
    deps = [set() for _ in range(len(obs))]
    for i, ob in enumerate(obs):
        for j, other_ob in enumerate(obs[i + 1:]):
            if not set(ob).isdisjoint(other_ob):
                deps[i].add(i)
                deps[i].add(j + i + 1)
                deps[j + i + 1].add(i)

    deps = tuple(tuple(dep) for dep in deps)

    return deps


def limits_overlap(spaces: ztyping.SpaceOrSpacesTypeInput, allow_exact_match: bool = False) -> bool:
    """Check if _any_ of the limits of `spaces` overlaps with _any_ other of `spaces`.

    This also checks multiple limits within one space. If `allow_exact_match` is set to true, then
    an *exact* overlap of limits is allowed.


    Args:
        spaces:
        allow_exact_match: An exact overlap of two limits is counted as "not overlapping".
            Example: limits from -1 to 3 and 4 to 5 to *NOT* overlap with the limits 4 to 5 *iff*
            `allow_exact_match` is True.

    Returns:
        If there are overlapping limits.
    """
    # TODO(Mayou36): add approx comparison global in zfit
    eps = 1e-8  # epsilon for float comparisons
    spaces = convert_to_container(spaces, container=tuple)
    all_obs = common_obs(spaces=spaces)
    for obs in all_obs:
        lowers = []
        uppers = []
        for space in spaces:
            if not space.has_limits or obs not in space.obs:
                continue
            else:
                index = space.obs.index(obs)

            for spa in space:
                lower, upper = spa.rect_limits  # TODO: new space
                low = lower[:, index]
                up = upper[:, index]

                for other_lower, other_upper in zip(lowers, uppers):
                    if allow_exact_match and np.allclose(other_lower, low) and np.allclose(other_upper, up):
                        continue
                    # TODO(Mayou36): tol? add global flags?
                    low_overlaps = np.all(other_lower - eps < low) and np.all(low < other_upper - eps)
                    up_overlaps = np.all(other_lower + eps < up) and np.all(up < other_upper + eps)
                    overlap = low_overlaps or up_overlaps
                    if overlap:
                        return True
                lowers.append(low)
                uppers.append(up)
    return False


def common_obs(spaces: ztyping.SpaceOrSpacesTypeInput) -> Union[List[str], bool]:
    """Extract the union of `obs` from `spaces` in the order of `spaces`.

    For example:
        | space1.obs: ['obs1', 'obs3']
        | space2.obs: ['obs2', 'obs3', 'obs1']
        | space3.obs: ['obs2']

        returns ['obs1', 'obs3', 'obs2']

    Args:
        spaces: :py:class:`~zfit.Space`s to extract the obs from

    Returns:
        The observables as `str` or False if not every space has observables
    """
    spaces = convert_to_container(spaces, container=tuple)
    all_obs = []
    for space in spaces:
        if space.obs is None:
            return False
        for ob in space.obs:
            if ob not in all_obs:
                all_obs.append(ob)
    return all_obs


def common_axes(spaces: ztyping.SpaceOrSpacesTypeInput) -> Union[List[str], bool]:
    """Extract the union of `axes` from `spaces` in the order of `spaces`.

    For example:
        | space1.axes: [1, 3]
        | space2.axes: [2, 3, 1]
        | space3.axes: [2]

        returns [1, 3, 2]

    Args:
        spaces: :py:class:`~zfit.Space`s to extract the axes from

    Returns:
        The axes as int or False if not every space has axes
    """
    spaces = convert_to_container(spaces, container=tuple)
    all_axes = []
    for space in spaces:
        if space.axes is None:
            return False
        for ax in space.axes:
            if ax not in all_axes:
                all_axes.append(ax)
    return all_axes


def obs_subsets(dimensionals: Iterable[ZfitDimensional]) -> Dict[Set[str], ZfitDimensional]:
    """Split `dimensionals` into the smallest subgroup of obs and return a dict.

    Args:
        dimensionals: An Iterable containing two or more ZfitDimensional that should be split into the smallest subset.

    Returns:
        Dict with the keys being sets of observables and the values, an iterable, containing the ZfitDimensional
    """
    obs_dims = {}
    for dim in dimensionals:
        for obs in obs_dims:
            if obs.intersection(dim.obs):
                union = obs.union(dim.obs)
                obs_dims[union] = obs_dims.pop(obs)
                obs_dims[union].append(dim)
                break  # we had a match, go to the next dim
        else:
            obs_dims[frozenset(dim.obs)] = [dim]

    return obs_dims



#  Copyright (c) 2021 zfit
import pytest
import tensorflow as tf

import zfit
from zfit import z
from zfit.core.coordinates import Coordinates
from zfit.core.space import ANY, Limit, Space
from zfit.util.exception import (CoordinatesUnderdefinedError,
                                 LimitsIncompatibleError)


@pytest.fixture(autouse=True, scope="module")
def setup_teardown_vectors():
    Limit._experimental_allow_vectors = True
    yield
    Limit._experimental_allow_vectors = False


obs1 = ('a', 'b', 'c', 'd', 'e')
obs2 = ('c', 'b', 'd', 'e', 'a')
axes1 = (0, 1, 2, 3, 4)
axes2 = (2, 1, 3, 4, 0)
coords1obs = Coordinates(obs1)
coords1 = Coordinates(obs1, axes1)
coords1mixed = Coordinates(obs1, axes2)
coords1axes = Coordinates(axes=axes1)
coords2obs = Coordinates(obs2)
coords2 = Coordinates(obs2, axes2)
coords2mixed = Coordinates(obs2, axes1)
coords2axes = Coordinates(axes=axes2)

limits1 = ([-2, -1, 0, 1, 2], [3, 4, 5, 6, 7])
limits2 = ([0, -1, 1, 2, -2], [5, 4, 6, 7, 3])

limits1limit = Limit(limits1, n_obs=5)
limits2limit = Limit(limits2, n_obs=5)

limits1space = Space(obs=obs1, axes=axes1, limits=limits1)
limits2space = Space(obs=obs2, axes=axes2, limits=limits2)

limit_fn = lambda x: x
limits1func = Space(obs=obs1, axes=axes1, limits=limit_fn, rect_limits=limits1)
limits2func = limits1func.with_obs(obs2)

limits3 = ([11, 12, 13, 14, 15], [31, 41, 51, 61, 71])
limits4 = ([13, 12, 14, 15, 11], [51, 41, 61, 71, 31])

limits1vector = ([[-21, -11, -91, 11, 21], [-22, -12, -92, 12, 22], [-23, -13, -93, 13, 23]],
                 [[31, 41, 51, 61, 71], [32, 42, 52, 62, 72], [33, 43, 53, 63, 73]])
limits2vector = ([[-91, -11, 11, 21, -21], [-92, -12, 12, 22, -22], [-93, -13, 13, 23, -23]],
                 [[51, 41, 61, 71, 31], [52, 42, 62, 72, 32], [53, 43, 63, 73, 33]])

limits1tf = (z.convert_to_tensor([-2, -1, 0, 1, 2]), z.convert_to_tensor([3, 4, 5, 6, 7]))
limits2tf = (z.convert_to_tensor([0, -1, 1, 2, -2]), z.convert_to_tensor([5, 4, 6, 7, 3]))

limits1mixed_any = ([-2, ANY, 0, ANY, 2], [3, 4, 5, ANY, 7])
limits2mixed_any = ([0, ANY, ANY, 2, -2], [5, 4, ANY, 7, 3])

limits1any = ([ANY] * 5, [ANY] * 5)
limits2any = ([ANY] * 5, [ANY] * 5)

limits_to_test = [
    [limits1, limits2],
    [limits1limit, limits2limit],
    [limits1space, limits2space],
    [limits1tf, limits2tf],
    [limits1mixed_any, limits2mixed_any],
    [limits1any, limits2any],
    [limits1vector, limits2vector],
    [{'multi': [limits1, limits3]}, {'multi': [limits2, limits4]}],
    [{'multi': [limits1any, limits2any]}, {'multi': [limits1any, limits2any]}],
    [{'multi': [limits1, limits2any]}, {'multi': [limits1any, limits2]}],
]


def test_illegal_bounds():
    with pytest.raises(tf.errors.InvalidArgumentError):
        _ = zfit.Space(['obs1', 'obs2'], ([-1, 4], [2, 3]))


def test_extract_limits():
    obs1 = ['a']
    space1 = Space('a', (0, 1))
    obs2 = ['b', 'c']
    limit2 = Limit(limit_fn=lambda x: x, rect_limits=([1, 2], [2, 3]), n_obs=2)
    obs3 = ['d', 'e', 'f']
    limits3_dict = {'obs': {ob: Limit((i, i + 10)) for i, ob in enumerate(obs3)}}
    space3 = Space(obs3, limits=limits3_dict)
    limits_dict = {'obs': {
        tuple(obs1): space1,
        tuple(obs2): limit2,
        tuple(obs3): space3,
    }
    }
    space = Space(obs1 + obs2 + obs3, limits_dict)

    extracted_limits = space.get_limits(obs1)['obs']
    assert list(extracted_limits.values())[0] == space1
    extracted_limits = space.get_limits(obs2)['obs']
    assert list(extracted_limits.values())[0] == limit2
    extracted_limits = space.get_limits(obs3)['obs']
    assert list(extracted_limits.values())[0] == space3

    extracted_limits = space.get_limits(obs3[0])
    # assert list(extracted_limits.values())[0] == limits3_dict[obs3[0]]
    # obs9 = obs3[0:2] + obs2
    # extracted_limits = space._extract_limits(obs9)
    # assert extracted_limits is limits3_dict[]


def test_rect_limits():
    obs1 = ['a']
    axes1 = [0]
    space1 = Space('a', (0, 1))
    space1_nolim = Space('a')
    assert not space1_nolim.has_limits
    assert space1.has_limits
    assert space1.has_rect_limits
    space1_lim = space1_nolim.with_limits((0, 1))
    space1_ax = Space(axes=0, limits=(0, 1))
    lower, upper = space1.rect_limits
    assert lower == 0
    assert upper == 1

    lower, upper = space1_ax.rect_limits
    assert lower == 0
    assert upper == 1

    lower, upper = space1_lim.rect_limits
    assert lower == 0
    assert upper == 1


def space_factory(*args, limits=None, **kwargs):
    """
    Args:
        *args:
        limits:
        **kwargs:
    """
    if isinstance(limits, dict):
        limit1, limit3 = limits['multi']
        space1 = zfit.Space(*args, limits=limit1, **kwargs)
        space3 = zfit.Space(*args, limits=limit3, **kwargs)
        return space1 + space3
    else:
        space = zfit.Space(*args, limits=limits, **kwargs)
        return space


@pytest.mark.parametrize('limits', limits_to_test)
def test_with_coords(limits):
    """
    Args:
        limits:
    """
    limits1, limits2 = limits

    space1obs = space_factory(obs1, limits=limits1)
    space1 = space_factory(obs1, limits=limits1, axes=axes1)
    space1axes = space_factory(limits=limits1, axes=axes1)
    # space1mixed = space_factory(obs1, limits=limits1, axes=axes2)

    space2obs = space_factory(obs2, limits=limits2)
    space2 = space_factory(obs2, limits=limits2, axes=axes2)
    space2axes = space_factory(limits=limits2, axes=axes2)

    # define which space to use in this tests
    space_used = space1obs

    space = space_used.with_coords(coords1obs)
    assert space == space1obs

    space = space_used.with_coords(coords2obs)
    assert space == space2obs

    coords = coords2
    space = space_used.with_coords(coords)
    assert space == space2

    with pytest.raises(CoordinatesUnderdefinedError):
        space = space_used.with_coords(coords2axes)

    space = space_used.with_coords(coords2mixed)
    assert space.obs == coords2mixed.obs
    assert space.axes == coords2mixed.axes

    # define which space to use in this tests
    space_used = space1

    space = space_used.with_coords(coords1obs)
    assert space == space1

    space = space_used.with_coords(coords2obs)
    assert space == space2

    space = space_used.with_coords(coords2)
    assert space == space2

    space = space_used.with_coords(coords2mixed)
    assert space.axes == coords2mixed.axes
    assert space.obs == coords2mixed.obs

    space = space_used.with_coords(coords1mixed)
    assert space.obs == coords1mixed.obs
    assert space.axes == coords1mixed.axes

    space = space_used.with_coords(coords2axes)
    assert space == space2

    # define which space to use in this tests
    space_used = space1axes

    space = space_used.with_coords(coords1axes)
    assert space == space1axes

    with pytest.raises(CoordinatesUnderdefinedError):
        space = space_used.with_coords(coords2obs)

    space = space_used.with_coords(coords2)
    assert space == space2

    space = space_used.with_coords(coords2axes)
    assert space == space2axes


@pytest.mark.parametrize('limits', limits_to_test)
def test_with_obs(limits):
    """
    Args:
        limits:
    """
    limits1, limits2 = limits

    space1obs = space_factory(obs1, limits=limits1)
    space1 = space_factory(obs1, limits=limits1, axes=axes1)
    space1axes = space_factory(limits=limits1, axes=axes1)
    using_space_as_lim = isinstance(limits1, Space)
    if isinstance(limits1, Space):
        limits1 = limits1.rect_limits
    space1mixed = space_factory(obs1, limits=limits1, axes=axes2)

    space2obs = space_factory(obs2, limits=limits2)
    space2 = space_factory(obs2, limits=limits2, axes=axes2)
    if isinstance(limits2, Space):
        limits2 = limits2.rect_limits
    space2axes = space_factory(limits=limits2, axes=axes2)
    space2mixed = space_factory(obs2, limits=limits1, axes=axes1)

    # define which space to use in this tests
    space_used = space1obs

    space = space_used.with_obs(obs1)
    assert space == space1obs

    space = space_used.with_obs(obs2)
    assert space == space2obs

    # define which space to use in this tests
    space_used = space1

    space = space_used.with_obs(obs1)
    assert space == space1

    space = space_used.with_obs(obs2)
    assert space == space2

    # define which space to use in this tests
    space_used = space1axes

    space = space_used.with_obs(obs1)
    assert space == space1

    space = space_used.with_obs(obs2)
    if using_space_as_lim:

        assert space == space2mixed
    else:
        assert space == space2mixed


@pytest.mark.parametrize('limits', limits_to_test)
def test_with_axes(limits):
    """
    Args:
        limits:
    """
    limits1, limits2 = limits

    space1obs = space_factory(obs1, limits=limits1)
    space1 = space_factory(obs1, limits=limits1, axes=axes1)
    space1axes = space_factory(limits=limits1, axes=axes1)
    using_space_as_lim = isinstance(limits1, Space)
    if isinstance(limits1, Space):
        limits1 = limits1.rect_limits
    space1mixed = space_factory(obs1, limits=limits1, axes=axes2)

    space2 = space_factory(obs2, limits=limits2, axes=axes2)
    if isinstance(limits2, Space):
        limits2 = limits2.rect_limits
    space2axes = space_factory(limits=limits2, axes=axes2)

    # define which space to use in this tests
    space_used = space1obs

    space = space_used.with_axes(axes1)
    assert space == space1

    space = space_used.with_axes(axes2)
    if using_space_as_lim:
        assert space == space1mixed
    else:
        assert space == space1mixed

    # define which space to use in this tests
    space_used = space1

    space = space_used.with_axes(axes1)
    assert space == space1

    space = space_used.with_axes(axes2)
    assert space == space2

    # define which space to use in this tests
    space_used = space1axes

    space = space_used.with_axes(axes1)
    assert space == space1axes

    space = space_used.with_axes(axes2)
    assert space == space2axes


@pytest.mark.parametrize('limits', limits_to_test)
def test_space_add(limits):
    """
    Args:
        limits:
    """
    limits1, limits2 = limits

    space1obs = space_factory(obs1, limits=limits1)
    space1 = space_factory(obs1, limits=limits1, axes=axes1)
    space1axes = space_factory(limits=limits1, axes=axes1)
    if isinstance(limits1, Space):
        limits1 = limits1.rect_limits
    space1mixed = space_factory(obs1, limits=limits1, axes=axes2)

    space2obs = space_factory(obs2, limits=limits2)
    space2 = space_factory(obs2, limits=limits2, axes=axes2)

    space2axes = space_factory(limits=limits2, axes=axes2)
    if isinstance(limits2, Space):
        limits2 = limits2.rect_limits
    space2mixed = space_factory(obs2, limits=limits2, axes=axes1)

    space = space1 + space2
    assert space.obs == obs1
    assert space.axes == axes1
    for spm, sp in zip(space, [*space1, *space2]):
        assert spm == sp

    space = space2 + space1
    assert space.obs == obs2
    assert space.axes == axes2
    for spm, sp in zip(space, [*space2, *space1]):
        assert spm == sp

    space = space1obs + space2
    assert space.obs == obs1
    assert space.axes == None

    space = space1axes + space2
    assert space.obs == None
    assert space.axes == axes1

    space = space2obs + space1
    assert space.obs == obs2
    assert space.axes == None

    space = space2axes + space1
    assert space.obs == None
    assert space.axes == axes2


def test_combine_spaces():
    shift = 30

    lower1, upper1 = [0, 1], [14, 13]
    lower1b, upper1b = [0 + shift, 1 + shift], [14 + shift, 13 + shift]
    lower2, upper2 = [-4, 1], [10, 13]
    lower2b, upper2b = [-4 + shift, 1 + shift], [10 + shift, 13 + shift]
    lower3, upper3 = [9, 1, 0], [11, 13, 14]
    obs1 = ['x', 'y']
    space1a = zfit.Space(obs1, limits=(lower1, upper1))
    space1b = zfit.Space(obs1, limits=(lower1b, upper1b))
    obs2 = ['z', 'y']
    space2a = zfit.Space(obs2, limits=(lower2, upper2))
    space2b = zfit.Space(obs2, limits=(lower2b, upper2b))
    obs3 = ['a', 'y', 'x']
    space3 = zfit.Space(obs3, limits=(lower3, upper3))
    space3inc = zfit.Space(obs3, limits=(lower3, upper3[::-1]))

    lower12 = [lower1[0], lower1[1], lower2[0]]
    upper12 = [upper1[0], upper1[1], upper2[0]]
    space12a = zfit.Space(('x', 'y', 'z'), limits=(lower12, upper12))
    space12b = zfit.Space(('x', 'y', 'z'), limits=([low + shift for low in lower12], [up + shift for up in upper12]))

    # space3 = zfit.Space(('x', 'y', 'z'), limits=([lower1[0], lower1[1], lower2[0]], [upper1[0], upper1[1], upper2[0]]))
    obs2inv = space2a.with_obs(['y', 'z'])

    space = space1a * space2a
    assert space == space12a
    assert space == space12a * obs2inv
    assert space == space12a * obs2inv * space12a

    space = space3 * space1a
    assert space == space3

    with pytest.raises(LimitsIncompatibleError):
        space1a * space3inc

    space12 = space12a + space12b
    space1 = space1a + space1b
    space2 = space2a + space2b
    space = space1 * space2
    assert space == space12



#  Copyright (c) 2021 zfit
import warnings
from typing import List, Mapping, Optional

import iminuit
import numpy as np

from ..core.interfaces import ZfitLoss
from ..core.parameter import Parameter, assign_values, set_values
from ..settings import run
from ..util.cache import GraphCachable
from ..util.deprecation import deprecated_args
from ..util.exception import MaximumIterationReached
from .baseminimizer import (BaseMinimizer, minimize_supports,
                            print_minimization_status)
from .fitresult import FitResult
from .strategy import ZfitStrategy
from .termination import EDM, ConvergenceCriterion


class Minuit(BaseMinimizer, GraphCachable):
    _DEFAULT_name = "Minuit"

    @deprecated_args(None, "Use `options` instead.", 'minimizer_options')
    @deprecated_args(None, "Use `maxiter` instead.", 'ncall')
    @deprecated_args(None, "Use `mode` instead.", 'minimize_strategy')
    @deprecated_args(None, "Use `gradient` instead.", 'minuit_grad')
    @deprecated_args(None, "Use `gradient` instead.", 'use_minuit_grad')
    def __init__(self,
                 tol: Optional[float] = None,
                 mode: Optional[int] = None,
                 gradient: Optional[bool] = None,
                 verbosity: Optional[int] = None,
                 options: Optional[Mapping[str, object]] = None,
                 maxiter: Optional[int] = None,
                 criterion: Optional[ConvergenceCriterion] = None,
                 strategy: Optional[ZfitStrategy] = None,
                 name: Optional[str] = None,
                 # legacy arguments
                 use_minuit_grad: Optional[bool] = None,
                 minuit_grad=None,
                 minimize_strategy=None,
                 ncall=None,
                 minimizer_options=None,
                 ):
        """Minuit is a longstanding and well proven algorithm of the L-BFGS-B class implemented in `iminuit`_.

        The package iminuit is the fast, interactive minimizer based on the Minuit2 C++ library; the latter is
        maintained by CERN’s ROOT team. It is an especially robust minimizer that finds the global minimum
        quiet reliably. It is however, like all local minimizers, still rather dependent on close enough
        initial values.

        .. _iminuit: https://iminuit.readthedocs.io/en/stable/.


        Args:
            tol:  |@doc:minimizer.tol| Termination value for the
                   convergence/stopping criterion of the algorithm
                   in order to determine if the minimum has
                   been found. Defaults to 1e-3. |@docend:minimizer.tol|
            mode: A number used by minuit to define the internal minimization strategy, either 0, 1 or 2.
                As `explained in the iminuit docs <>`_, they mean:
                - 0 (default with zfit gradient): the fastest and the number of function calls required to minimise
                    scales linearly with the number of fitted parameters. The Hesse matrix is not computed during the
                    minimisation (only an approximation that is continuously updated).
                    When the number of fitted parameters > 10, you should prefer this strategy.
                - 1 (default with Minuit gradient) medium in speed. The number of function calls required
                    scales quadratically with the number of fitted parameters. The different scales comes from the fact
                     that the Hesse matrix is explicitly computed in a Newton step,
                     if Minuit detects significant correlations between parameters.
                - 2: same quadratic scaling as strategy 1 but is even slower. The Hesse matrix is
                    always explicitly computed in each Newton step.
            gradient: If True, iminuit uses its internal numerical gradient calculation instead of the
                (analytic/numerical) gradient provided by TensorFlow/zfit. If False or ``'zfit'``, the latter
                is used. For smaller datasets with less stable losses, the internal Minuit gradient often performs
                better while the zfit provided gradient improves the convergence rate for larger (10'000+) datasets.
            verbosity: |@doc:minimizer.verbosity| Verbosity of the minimizer. Has to be between 0 and 10.
              The verbosity has the meaning:

               - a value of 0 means quiet and no output
               - above 0 up to 5, information that is good to know but without
                 flooding the user, corresponding to a "INFO" level.
               - A value above 5 starts printing out considerably more and
                 is used more for debugging purposes.
               - Setting the verbosity to 10 will print out every
                 evaluation of the loss function and gradient.

               Some minimizer offer additional output which is also
               distributed as above but may duplicate certain printed values. |@docend:minimizer.verbosity| This
                also changes the iminuit internal verbosity at around 7.
            options: Additional options that will be directly passsed into :meth:`~iminuitMinuit.migrad`
            maxiter: |@doc:minimizer.maxiter| Approximate number of iterations.
                   This corresponds to roughly the maximum number of
                   evaluations of the `value`, 'gradient` or `hessian`. |@docend:minimizer.maxiter|
            criterion: |@doc:minimizer.criterion| Criterion of the minimum. This is an
                   estimated measure for the distance to the
                   minimum and can include the relative
                   or absolute changes of the parameters,
                   function value, gradients and more.
                   If the value of the criterion is smaller
                   than ``loss.errordef * tol``, the algorithm
                   stopps and it is assumed that the minimum
                   has been found. |@docend:minimizer.criterion|
            strategy: |@doc:minimizer.strategy| A class of type `ZfitStrategy` that takes no
                   input arguments in the init. Determines the behavior of the minimizer in
                   certain situations, most notably when encountering
                   NaNs. It can also implement a callback function. |@docend:minimizer.strategy|
            name: |@doc:minimizer.name| Human readable name of the minimizer. |@docend:minimizer.name|


            use_minuit_grad: deprecated, legacy.
            minuit_grad: deprecated, legacy.
            minimize_strategy: deprecated, legacy.
            ncall: deprecated, legacy.
            minimizer_options: deprecated, legacy.
        """
        # legacy
        if isinstance(mode, float) or isinstance(tol, int):
            raise TypeError('mode has to be int, tol a float. The API changed, make sure you use the'
                            ' right parameters.')
        if minimizer_options is not None:
            options = minimizer_options
        if ncall is not None:
            maxiter = ncall
        if minimize_strategy is not None:
            mode = minimize_strategy
        use_grad_legacy = use_minuit_grad if use_minuit_grad is not None else minuit_grad
        if use_grad_legacy is not None:
            gradient = use_grad_legacy
        # end legacy

        if gradient == 'zfit':
            gradient = False
        gradient = True if gradient is None else gradient

        self._internal_maxiter = 20

        options = {} if options is None else options
        options['ncall'] = 0 if maxiter is None else maxiter
        if mode is None:
            mode = 0 if not gradient else 1
        else:
            mode = mode
        if mode not in range(3):
            raise ValueError(f"mode has to be 0, 1 or 2, not {mode}.")
        options['strategy'] = mode

        super().__init__(name=name, strategy=strategy, tol=tol, verbosity=verbosity, criterion=criterion,
                         maxiter=1e20,
                         minimizer_options=options)
        self._minuit_minimizer = None
        self._use_tfgrad_internal = not gradient
        self.minuit_grad = gradient

    # TODO 0.7: legacy, remove `_use_tfgrad`
    @property
    def _use_tfgrad(self):
        warnings.warn("Do not use `minimizer._use_tfgrad`, this will be removed. Use `minuit_grad` instead in the"
                      " initialization.", stacklevel=2)
        return self._use_tfgrad_internal

    @minimize_supports()
    def _minimize(self, loss: ZfitLoss, params: List[Parameter], init):
        if init:
            assign_values(params=params, values=init)
        criterion = self.create_criterion(loss, params)

        minimizer, minimize_options, evaluator = self._make_minuit(loss, params, init)

        self._minuit_minimizer = minimizer

        valid = False
        message = ""
        maxiter_reached = False
        for i in range(self._internal_maxiter):

            # perform minimization
            try:
                minimizer = minimizer.migrad(**minimize_options)
            except MaximumIterationReached as error:
                if minimizer is None:  # it didn't even run once
                    raise MaximumIterationReached("Maximum iteration reached on first wrapped minimizer call. This"
                                                  "is likely to a too low number of maximum iterations (currently"
                                                  f" {evaluator.maxiter}) or wrong internal tolerances, in which"
                                                  f" case: please fill an issue on github.") from error
                maxiter_reached = True
                message = "Maxiter reached"
            else:
                if evaluator.maxiter is not None:
                    maxiter_reached = evaluator.niter > evaluator.maxiter
            if type(criterion) == EDM:  # use iminuits edm
                criterion.last_value = minimizer.fmin.edm
                converged = not minimizer.fmin.is_above_max_edm
            else:
                fitresult = FitResult.from_minuit(loss=loss, params=params, minuit=minimizer, minimizer=self,
                                                  valid=valid, message=message)
                converged = criterion.converged(fitresult)

            if self.verbosity > 5:
                internal_tol = {'edm_minuit': minimizer.fmin.edm}

                print_minimization_status(converged=converged,
                                          criterion=criterion,
                                          evaluator=evaluator,
                                          i=i,
                                          fmin=minimizer.fval,
                                          internal_tol=internal_tol)

            if converged or maxiter_reached:
                assign_values(params, minimizer.values)  # make sure it's at the right value
                if not maxiter_reached:
                    valid = True
                break

        fitresult = FitResult.from_minuit(loss=loss,
                                          params=params,
                                          criterion=criterion,
                                          minuit=minimizer,
                                          minimizer=self.copy(),
                                          valid=valid,
                                          message=message)
        return fitresult

    def _make_minuit(self, loss, params, init):

        evaluator = self.create_evaluator(loss, params)

        # create options
        minimizer_options = self.minimizer_options.copy()
        minimize_options = {}
        precision = minimizer_options.pop('precision', None)
        minimize_options['ncall'] = minimizer_options.pop('ncall')
        minimizer_init = {}
        if 'errordef' in minimizer_options:
            raise ValueError("errordef cannot be specified for Minuit as this is already defined in the Loss.")
        loss_errordef = loss.errordef
        if not isinstance(loss_errordef, (float, int)):
            raise ValueError("errordef has to be a float")
        minimizer_init['errordef'] = loss_errordef
        minimizer_init['pedantic'] = minimizer_options.pop('pedantic', False)
        minimizer_setter = {}
        minimizer_setter['strategy'] = minimizer_options.pop('strategy')
        if self.verbosity > 8:
            minuit_verbosity = 3
        elif self.verbosity > 6:
            minuit_verbosity = 2
        elif self.verbosity > 1:
            minuit_verbosity = 1
        else:
            minuit_verbosity = 0
        if minimizer_options:
            raise ValueError(f"The following options are not (yet) supported: {minimizer_options}")
        init_values = np.array(run(params))

        # create Minuit compatible names
        params_name = [param.name for param in params]
        # TODO 0.7: legacy, remove `_use_tfgrad`
        grad_func = evaluator.gradient if self._use_tfgrad_internal or not self.minuit_grad else None
        minimizer = iminuit.Minuit(evaluator.value, init_values,
                                   grad=grad_func,
                                   name=params_name,
                                   )
        minimizer.precision = precision
        approx_step_sizes = {}
        # get possible initial step size from previous minimizer
        if init:
            approx_step_sizes = init.hesse(params=params, method='approx')

        empty_dict = {}
        for param in params:
            step_size = approx_step_sizes.get(param, empty_dict).get('error')
            if step_size is None and param.has_step_size:
                step_size = param.step_size
            if step_size is not None:
                minimizer.errors[param.name] = step_size
        # set limits
        for param in params:
            if param.has_limits:
                minimizer.limits[param.name] = (param.lower, param.upper)
        # set options
        minimizer.errordef = loss.errordef
        minimizer.print_level = minuit_verbosity
        strategy = minimizer_setter.pop('strategy')
        minimizer.strategy = strategy
        minimizer.tol = (self.tol / 0.002  # iminuit multiplies by default with 0.002
                         / loss.errordef)  # to account for the loss
        assert not minimizer_setter, "minimizer_setter is not empty, bug. Please report. minimizer_setter: {}".format(
            minimizer_setter)
        return minimizer, minimize_options, evaluator

    def copy(self):
        tmp_minimizer = self._minuit_minimizer
        new_minimizer = super().copy()
        new_minimizer._minuit_minimizer = tmp_minimizer
        return new_minimizer
